/*
* TWI_Slave.c
*
* Created: 04.02.2015 11:27:14
*  Author: Stanislav Chystiakov
*/
/*****************************************************************************
*
* Atmel Corporation
*
* File : TWI_Slave.c
* Compiler : IAR EWAAVR 2.28a/3.10c
* Revision : $Revision: 2475 $
* Date : $Date: 2007-09-20 12:00:43 +0200 (to, 20 sep 2007) $
* Updated by : $Author: mlarsson $
*
* Support mail : avr@atmel.com
*
* Supported devices : All devices with a TWI module can be used.
* The example is written for the ATmega16
*
* AppNote : AVR311 - TWI Slave Implementation
*
* Description : This is sample driver to AVRs TWI module.
* It is interupt driveren. All functionality is controlled through
* passing information to and from functions. Se main.c for samples
* of how to use the driver.
*
****************************************************************************/

#include "TWI_slave.h"
#include "port_init.h"
#include "timer.h"
#include "ADC.h"
#include <util/delay.h>
static unsigned char TWI_rx_buf[TWI_BUFFER_SIZE]; // Transceiver buffer. Set the size in the header file
static unsigned char TWI_tx_buf[TWI_BUFFER_SIZE]; 
static unsigned char TWI_msgSize = 0; // Number of bytes to be transmitted.
static unsigned char TWI_state = TWI_NO_STATE; // State byte. Default set to TWI_NO_STATE.
// This is true when the TWI is in the middle of a transfer
// and set to false when all bytes have been transmitted/received
// Also used to determine how deep we can sleep.
static unsigned char TWI_busy = 0;
volatile union TWI_statusReg_t TWI_statusReg = {0}; // TWI_statusReg is defined in TWI_Slave.h
	
int adc_buffer;
int adcDataBuf[3] = {0, 0, 0};
	
int dataToSend[3] = {1000, 1001, 1002}; // Test Only<!
extern int countsToError;
unsigned char messageBuf_rx[TWI_BUFFER_SIZE] = {0, 0, 0, 0};
unsigned char messageBuf_tx[TWI_BUFFER_SIZE] = {0, 0, 0, 0};

// When there has been an error, this function is run and takes care of it
unsigned char TWI_Act_On_Failure_In_Last_Transmission ( unsigned char TWIerrorMsg );	
	
	
	
/****************************************************************************
Call this function to set up the TWI slave to its initial standby state.
Remember to enable interrupts from the main application after initializing the TWI.
Pass both the slave address and the requrements for triggering on a general call in the
same byte. Use e.g. this notation when calling this function:
TWI_Slave_Initialise( (TWI_slaveAddress<<TWI_ADR_BITS) | (TRUE<<TWI_GEN_BIT) );
The TWI module is configured to NACK on any requests. Use a TWI_Start_Transceiver function to
start the TWI.
****************************************************************************/
void TWI_Slave_Initialise( unsigned char TWI_ownAddress )
{
	cli();
	TWAR = TWI_ownAddress; // Set own TWI slave address. Accept TWI General Calls.
	TWCR = (1<<TWEN)| // Enable TWI-interface and release TWI pins.
	(0<<TWIE)|(0<<TWINT)| // Disable TWI Interupt.
	(0<<TWEA)|(0<<TWSTA)|(0<<TWSTO)| // Do not ACK on any requests, yet.
	(0<<TWWC); //
	TWI_busy = 0;
	sei();
}
/****************************************************************************
Call this function to test if the TWI_ISR is busy transmitting.
****************************************************************************/
unsigned char TWI_Transceiver_Busy( void )
{
	return TWI_busy;
}

unsigned char TWI_Transceiver_Busy_Wait(void)
{
	unsigned int errcnt=TWI_ERR_TIMEOUT;
	while (TWI_Transceiver_Busy()) // Wait until TWI has completed the transmission.
	{ 
		//_delay_us(1); 
	    if (errcnt--==0) return TWI_BUS_ERROR; //timeout
	}
}

/****************************************************************************
Call this function to fetch the state information of the previous operation. The function will hold execution (loop)
until the TWI_ISR has completed with the previous operation. If there was an error, then the function
will return the TWI State code.
****************************************************************************/
unsigned char TWI_Get_State_Info( void )
{
	TWI_Transceiver_Busy_Wait(); // Wait until TWI has completed the transmission.
	return ( TWI_state ); // Return error state.
}
/****************************************************************************
Call this function to send a prepared message, or start the Transceiver for reception. Include
a pointer to the data to be sent if a SLA+W is received. The data will be copied to the TWI buffer.
Also include how many bytes that should be sent. Note that unlike the similar Master function, the
Address byte is not included in the message buffers.
The function will hold execution (loop) until the TWI_ISR has completed with the previous operation,
then initialize the next operation and return.
****************************************************************************/
void TWI_Start_Transceiver_With_Data( unsigned char *msg, unsigned char msgSize )
{
	unsigned char temp;
	TWI_Transceiver_Busy_Wait(); // Wait until TWI has completed the transmission.
	TWI_msgSize = msgSize; // Number of data to transmit.
	for ( temp = 0; temp < msgSize; temp++ ) // Copy data that may be transmitted if the TWI Master requests data.
	{
		TWI_tx_buf[ temp ] = msg[ temp ];
	}
	TWI_statusReg.all = 0;
	TWI_state = TWI_NO_STATE ;
	TWCR = (1<<TWEN)| // TWI Interface enabled.
		(1<<TWIE)|(1<<TWINT)| // Enable TWI Interrupt and clear the flag.
		(1<<TWEA)|(0<<TWSTA)|(0<<TWSTO)| // Prepare to ACK next time the Slave is addressed.
		(0<<TWWC); //
	TWI_busy = 1;
}
/****************************************************************************
Call this function to start the Transceiver without specifying new transmission data. Useful for restarting
a transmission, or just starting the transceiver for reception. The driver will reuse the data previously put
in the transceiver buffers. The function will hold execution (loop) until the TWI_ISR has completed with the
previous operation, then initialize the next operation and return.
****************************************************************************/
void TWI_Start_Transceiver( void )
{
	TWI_Transceiver_Busy_Wait(); // Wait until TWI has completed the transmission.
	TWI_statusReg.all = 0;
	TWI_state = TWI_NO_STATE ;
	TWCR = (1<<TWEN)| // TWI Interface enabled.
	(1<<TWIE)|(1<<TWINT)| // Enable TWI Interrupt and clear the flag.
	(1<<TWEA)|(0<<TWSTA)|(0<<TWSTO)| // Prepare to ACK next time the Slave is addressed.
	(0<<TWWC); //
	TWI_busy = 0;
}
/****************************************************************************
Call this function to read out the received data from the TWI transceiver buffer. I.e. first call
TWI_Start_Transceiver to get the TWI Transceiver to fetch data. Then Run this function to collect the
data when they have arrived. Include a pointer to where to place the data and the number of bytes
to fetch in the function call. The function will hold execution (loop) until the TWI_ISR has completed
with the previous operation, before reading out the data and returning.
If there was an error in the previous transmission the function will return the TWI State code.
****************************************************************************/
unsigned char TWI_Get_Data_From_Transceiver( unsigned char *msg, unsigned char msgSize )
{ // Data has been received from MASTER!!!
	unsigned char i;
	TWI_Transceiver_Busy_Wait(); // Wait until TWI has completed the transmission.
	if( TWI_statusReg.lastTransOK ) // Last transmission completed successfully.
	{
		for ( i=0; i<msgSize; i++ ) // Copy data from Transceiver buffer.
		{
			msg[ i ] = TWI_rx_buf[ i ];
		}
		TWI_statusReg.RxDataInBuf = FALSE; // Slave Receive data has been read from buffer.
	}
	return( TWI_statusReg.lastTransOK );
}
// ********** Interrupt Handlers ********** //
/****************************************************************************
This function is the Interrupt Service Routine (ISR), and called when the TWI interrupt is triggered;
that is whenever a TWI event has occurred. This function should not be called directly from the main
application.
****************************************************************************/
ISR(TWI_vect)
{
	static unsigned char TWI_rx_bufPtr;
	static unsigned char TWI_tx_bufPtr;
	switch (TWSR & TWI_TWSR_UNUSED_MARK)
	{
		case TWI_STX_ADR_ACK: // Own SLA+R has been received; ACK has been returned
			TWI_statusReg.transmitEnable = TRUE;
			// case TWI_STX_ADR_ACK_M_ARB_LOST: // Arbitration lost in SLA+R/W as Master; own SLA+R has been received; ACK has been returned
			TWI_tx_bufPtr = 0; // Set buffer pointer to first data location
		case TWI_STX_DATA_ACK: // Data byte in TWDR has been transmitted; ACK has been received
			TWDR = TWI_tx_buf[TWI_tx_bufPtr];
			TWI_tx_bufPtr++;
			TWCR = (1<<TWEN)| // TWI Interface enabled
			(1<<TWIE)|(1<<TWINT)| // Enable TWI Interupt and clear the flag to send byte
			(1<<TWEA)|(0<<TWSTA)|(0<<TWSTO)| //
			(0<<TWWC); //
			TWI_busy = 1;
			break;
		case TWI_STX_DATA_NACK: // Data byte in TWDR has been transmitted; NACK has been received.
			// I.e. this could be the end of the transmission.
			if (TWI_tx_bufPtr == TWI_msgSize) // Have we transceived all expected data?
			{
				TWI_statusReg.lastTransOK = TRUE; // Set status bits to completed successfully.
			}
			else // Master has sent a NACK before all data where sent.
			{
				TWI_state = TWSR; // Store TWI State as error message.
			}
			TWCR = (1<<TWEN)| // Enable TWI-interface and release TWI pins
			(1<<TWIE)|(1<<TWINT)| // Keep interrupt enabled and clear the flag
			(1<<TWEA)|(0<<TWSTA)|(0<<TWSTO)| // Answer on next address match
			(0<<TWWC); //
			TWI_busy = 0; // Transmit is finished, we are not busy anymore
			break;
		case TWI_SRX_GEN_ACK: // General call address has been received; ACK has been returned
			// case TWI_SRX_GEN_ACK_M_ARB_LOST: // Arbitration lost in SLA+R/W as Master; General call address has been received; ACK has been returned
			TWI_statusReg.genAddressCall = TRUE;
		case TWI_SRX_ADR_ACK: // Own SLA+W has been received ACK has been returned
			// case TWI_SRX_ADR_ACK_M_ARB_LOST: // Arbitration lost in SLA+R/W as Master; own SLA+W has been received; ACK has been returned
			// Dont need to clear TWI_S_statusRegister.generalAddressCall due to that it is the default state.
			TWI_statusReg.RxDataInBuf = TRUE;
			TWI_rx_bufPtr = 0; // Set buffer pointer to first data location
			// Reset the TWI Interrupt to wait for a new event.
			TWCR = (1<<TWEN)| // TWI Interface enabled
			(1<<TWIE)|(1<<TWINT)| // Enable TWI Interrupt and clear the flag to send byte
			(1<<TWEA)|(0<<TWSTA)|(0<<TWSTO)| // Expect ACK on this transmission
			(0<<TWWC);
			TWI_busy = 1;
			break;
		case TWI_SRX_ADR_DATA_ACK: // Previously addressed with own SLA+W; data has been received; ACK has been returned
		case TWI_SRX_GEN_DATA_ACK: // Previously addressed with general call; data has been received; ACK has been returned
			TWI_rx_buf[TWI_rx_bufPtr++] = TWDR;
			TWI_statusReg.lastTransOK = TRUE; // Set flag transmission successful.
			// Reset the TWI Interrupt to wait for a new event.
			TWCR = (1<<TWEN)| // TWI Interface enabled
			(1<<TWIE)|(1<<TWINT)| // Enable TWI Interrupt and clear the flag to send byte
			(1<<TWEA)|(0<<TWSTA)|(0<<TWSTO)| // Send ACK after next reception
			(0<<TWWC); //
			TWI_busy = 1;
			break;
		case TWI_SRX_STOP_RESTART: // A STOP condition or repeated START condition has been received while still addressed as Slave
			// Enter not addressed mode and listen to address match
			TWCR = (1<<TWEN)| // Enable TWI-interface and release TWI pins
			(1<<TWIE)|(1<<TWINT)| // Enable interrupt and clear the flag
			(1<<TWEA)|(0<<TWSTA)|(0<<TWSTO)| // Wait for new address match
			(0<<TWWC); //
			TWI_busy = 0; // We are waiting for a new address match, so we are not busy
			break;
		case TWI_SRX_ADR_DATA_NACK: // Previously addressed with own SLA+W; data has been received; NOT ACK has been returned
		case TWI_SRX_GEN_DATA_NACK: // Previously addressed with general call; data has been received; NOT ACK has been returned
		case TWI_STX_DATA_ACK_LAST_BYTE: // Last data byte in TWDR has been transmitted (TWEA = ?0?); ACK has been received
		// case TWI_NO_STATE // No relevant state information available; TWINT = ?0?
		case TWI_BUS_ERROR: // Bus error due to an illegal START or STOP condition
			TWI_state = TWSR; //Store TWI State as errormessage, operation also clears noErrors bit
			TWCR = (1<<TWSTO)|(1<<TWINT); //Recover from TWI_BUS_ERROR, this will release the SDA and SCL pins thus enabling other devices to use the bus
			break;
		default:
			TWI_state = TWSR; // Store TWI State as errormessage, operation also clears the Success bit.
			TWCR = (1<<TWEN)| // Enable TWI-interface and release TWI pins
			(1<<TWIE)|(1<<TWINT)| // Keep interrupt enabled and clear the flag
			(1<<TWEA)|(0<<TWSTA)|(0<<TWSTO)| // Acknowledge on any new requests.
			(0<<TWWC); //
			TWI_busy = 0; // Unknown status, so we wait for a new address match that might be something we can handle
	}
	
	if (!TWI_Transceiver_Busy() )
	{
		// Check if the last operation was successful
		if ( /*1*/ TWI_statusReg.lastTransOK/**/ )
		{
			// Check if the last operation was a reception
			if ( TWI_statusReg.RxDataInBuf )
			{
				TWI_Get_Data_From_Transceiver(messageBuf_rx, 3);
				switch (messageBuf_rx[0]) {
					case CMD_TWI_CHECK_PRESENCE: // read version and revision
					{
						messageBuf_rx[0] = 0;
						messageBuf_tx[0] = FREDFAN_VER;
						messageBuf_tx[1] = FREDFAN_REV;
						TWI_Start_Transceiver_With_Data( messageBuf_tx, 2 );
						break;
					}				
					case CMD_TWI_READ_VAL: // read constant values from uC (only for test!!)
					{
						messageBuf_rx[0] = 0;
						unsigned char i;
						for (i = 0; i < 3; i++) {
							messageBuf_tx[i] = (dataToSend[i]++ & 0xFF);
						}
						messageBuf_tx[3] =  ((dataToSend[0] >> 2) & 0xC0) | ((dataToSend[1] >> 4) & 0x30) | ((dataToSend[2] >> 6) & 0x0C);
						TWI_Start_Transceiver_With_Data(messageBuf_tx, 4 );
						break;
					}
					case CMD_TWI_READ_ADC_ALL: // calculate average for all channels (calculated by ATmega88PA)
					{
						unsigned char i;
						for (i = 0; i < 3; i++) {
						messageBuf_tx[i] = (adcDataBuf[i] & 0xFF);
						}
						messageBuf_tx[3] =  ((adcDataBuf[0] >> 2) & 0xC0) | ((adcDataBuf[1] >> 4) & 0x30) | ((adcDataBuf[2] >> 6) & 0x0C);
						TWI_Start_Transceiver_With_Data(messageBuf_tx, 4 );
						break;
					}
					case CMD_TWI_SET_AVERAGE:
					{
						setAverage(messageBuf_rx[1]);
						break;
					}
					case CMD_TWI_SET_PORT_DIR:
					{
						setPortDirection(messageBuf_rx[1], messageBuf_rx[2]);
						break;
					}
					case CMD_TWI_SET_PORT_LEV:
					{
						setPortLevel(messageBuf_rx[1], messageBuf_rx[2]);
						break;
					}
					case CMD_TWI_READ_PORT_LEV:
					{
						messageBuf_tx[0] = readPortLevel(messageBuf_rx[1]);
						TWI_Start_Transceiver_With_Data(messageBuf_tx, 1 );
						break;						
					}
					case CMD_RESET_ERR_TIME:
					{
						countsToError = COUNTS_TO_ERR;
						break;
					}
					//default:				
				}
			}
			else // Ends up here if the last operation was a transmission
			{
				//NOP(); // Put own code here.
			}
			// Check if the TWI Transceiver has already been started.
			// If not then restart it to prepare it for new receptions.
			//if ( ! TWI_Transceiver_Busy() )
			//{
			//	TWI_Start_Transceiver();
			//}
		}
		else // Ends up here if the last operation completed unsuccessfully
		{
			TWI_Act_On_Failure_In_Last_Transmission( TWI_Get_State_Info() );
		}
	}
}

unsigned char TWI_Act_On_Failure_In_Last_Transmission ( unsigned char TWIerrorMsg )
{
	// A failure has occurred, use TWIerrorMsg to determine the nature of the failure
	// and take appropriate actions.
	// Se header file for a list of possible failures messages.
	// This very simple example puts the error code on PORTB and restarts the transceiver with
	// all the same data in the transmission buffers.
	
	TWI_Start_Transceiver();
	return TWIerrorMsg;
}