/*
* TWI_Slave.h
*
* Created: 04.02.2015 11:30:13
*  Author:
*/


/*****************************************************************************
*
* Atmel Corporation
*
* File : TWI_Slave.h
* Compiler : IAR EWAAVR 2.28a/3.10c
* Revision : $Revision: 2475 $
* Date : $Date: 2007-09-20 12:00:43 +0200 (to, 20 sep 2007) $
* Updated by : $Author: mlarsson $
*
* Support mail : avr@atmel.com
*
* Supported devices : All devices with a TWI module can be used.
* The example is written for the ATmega16
*
* AppNote : AVR311 - TWI Slave Implementation
*
* Description : Header file for TWI_slave.c
* Include this file in the application.
*
****************************************************************************/
/*! \page MISRA
*
* General disabling of MISRA rules:
* * (MISRA C rule 1) compiler is configured to allow extensions
* * (MISRA C rule 111) bit fields shall only be defined to be of type unsigned int or signed int
* * (MISRA C rule 37) bitwise operations shall not be performed on signed integer types
* As it does not work well with 8bit architecture and/or IAR
* Other disabled MISRA rules
* * (MISRA C rule 109) use of union - overlapping storage shall not be used
* * (MISRA C rule 61) every non-empty case clause in a switch statement shall be terminated with a break statement
*/

#include <avr/interrupt.h>
#include <avr/io.h>


/****************************************************************************
TWI Status/Control register definitions
****************************************************************************/
#define TWI_SLV_ADDR				0x0C	// TWI slave address
#define TWI_BUFFER_SIZE				4		// Reserves memory for the drivers transceiver buffer.
											// Set this to the largest message size that will be sent including address byte.
#define TWI_ERR_TIMEOUT				1000	// s until Timeout is reported
#define COUNTS_TO_ERR				8		// (2sec * Fcpu)/(UPDATE_TIME * prescaler) [UPDATE_TIME from timer.h]

/***************************************************************************
Sample TWI transmission commands
***************************************************************************/
#define FREDFAN_VER					0x01	// FRED Fan board version
#define FREDFAN_REV					0x00	// FRED Fan board revision

#define TWI_CMD_MASTER_WRITE		0x10
#define TWI_TWSR_UNUSED_MARK        0xF8
//#define CMD_TWI_READ_ADC		 	0x20	// read only 1 ADC channel
//#define CMD_TWI_READ_ADC_AVRG		0x21	// read average only for 1 ADC channel (calculated by ATmega88PA)
#define CMD_TWI_READ_ADC_ALL		0x22	// read average for all channels (calculated by ATmega88PA)
#define CMD_TWI_READ_VAL			0x23	// read value from ATmega88PA (Test ONLY!!!)
#define CMD_TWI_CHECK_PRESENCE		0x24	// read FRED Fan board version and revision
#define CMD_TWI_SET_AVERAGE			0x25	// set number of ADC samples
#define CMD_TWI_SET_PORT_DIR		0x26	// set port direction /setPortDirection()
#define CMD_TWI_SET_PORT_LEV		0x27	// set port level /setPortLevel()
#define CMD_TWI_READ_PORT_LEV		0x28	// read port level /readPortLevel()


#define CMD_RESET_ERR_TIME			0x30	// reset error time

/****************************************************************************
Global definitions
****************************************************************************/
union TWI_statusReg_t // Status byte holding flags.
{
	unsigned char all;
	struct
	{
		unsigned char lastTransOK:1;
		unsigned char RxDataInBuf:1;
		unsigned char genAddressCall:1; // TRUE = General call, FALSE = TWI Address;
		unsigned char transmitEnable:1;
		unsigned char unusedBits:5;
	};
};
extern volatile union TWI_statusReg_t TWI_statusReg;

/****************************************************************************
Function definitions
****************************************************************************/
void TWI_Slave_Initialise( unsigned char );
unsigned char TWI_Transceiver_Busy( void );
unsigned char TWI_Get_State_Info( void );
void TWI_Start_Transceiver_With_Data( unsigned char * , unsigned char );
void TWI_Start_Transceiver( void );
unsigned char TWI_Get_Data_From_Transceiver( unsigned char *, unsigned char );
unsigned char TWI_Transceiver_Busy_Wait(void);
/****************************************************************************
Bit and byte definitions
****************************************************************************/
#define TWI_READ_BIT 0 // Bit position for R/W bit in "address byte".
#define TWI_ADR_BITS 1 // Bit position for LSB of the slave address bits in the init byte.
#define TWI_GEN_BIT 0 // Bit position for LSB of the general call bit in the init byte.
#define TRUE 1
#define FALSE 0
/****************************************************************************
TWI State codes
****************************************************************************/
// General TWI Master staus codes
#define TWI_START 0x08 // START has been transmitted
#define TWI_REP_START 0x10 // Repeated START has been transmitted
#define TWI_ARB_LOST 0x38 // Arbitration lost
// TWI Master Transmitter status codes
#define TWI_MTX_ADR_ACK 0x18 // SLA+W has been transmitted and ACK received
#define TWI_MTX_ADR_NACK 0x20 // SLA+W has been transmitted and NACK received
#define TWI_MTX_DATA_ACK 0x28 // Data byte has been transmitted and ACK received
#define TWI_MTX_DATA_NACK 0x30 // Data byte has been transmitted and NACK received
// TWI Master Receiver status codes
#define TWI_MRX_ADR_ACK 0x40 // SLA+R has been transmitted and ACK received
#define TWI_MRX_ADR_NACK 0x48 // SLA+R has been transmitted and NACK received
#define TWI_MRX_DATA_ACK 0x50 // Data byte has been received and ACK transmitted
#define TWI_MRX_DATA_NACK 0x58 // Data byte has been received and NACK transmitted
// TWI Slave Transmitter status codes
#define TWI_STX_ADR_ACK 0xA8 // Own SLA+R has been received; ACK has been returned
#define TWI_STX_ADR_ACK_M_ARB_LOST 0xB0 // Arbitration lost in SLA+R/W as Master; own SLA+R has been received; ACK has been returned
#define TWI_STX_DATA_ACK 0xB8 // Data byte in TWDR has been transmitted; ACK has been received
#define TWI_STX_DATA_NACK 0xC0 // Data byte in TWDR has been transmitted; NOT ACK has been received
#define TWI_STX_DATA_ACK_LAST_BYTE 0xC8 // Last data byte in TWDR has been transmitted (TWEA = 0); ACK has been received
// TWI Slave Receiver status codes
#define TWI_SRX_ADR_ACK 0x60 // Own SLA+W has been received ACK has been returned
#define TWI_SRX_ADR_ACK_M_ARB_LOST 0x68 // Arbitration lost in SLA+R/W as Master; own SLA+W has been received; ACK has been returned
#define TWI_SRX_GEN_ACK 0x70 // General call address has been received; ACK has been returned
#define TWI_SRX_GEN_ACK_M_ARB_LOST 0x78 // Arbitration lost in SLA+R/W as Master; General call address has been received; ACK has been returned
#define TWI_SRX_ADR_DATA_ACK 0x80 // Previously addressed with own SLA+W; data has been received; ACK has been returned
#define TWI_SRX_ADR_DATA_NACK 0x88 // Previously addressed with own SLA+W; data has been received; NOT ACK has been returned
#define TWI_SRX_GEN_DATA_ACK 0x90 // Previously addressed with general call; data has been received; ACK has been returned
#define TWI_SRX_GEN_DATA_NACK 0x98 // Previously addressed with general call; data has been received; NOT ACK has been returned
#define TWI_SRX_STOP_RESTART 0xA0 // A STOP condition or repeated START condition has been received while still addressed as Slave
// TWI Miscellaneous status codes
#define TWI_NO_STATE 0xF8 // No relevant state information available; TWINT = 0
#define TWI_BUS_ERROR 0x00 // Bus error due to an illegal START or STOP condition