/**
 * @file i2c_lib.C
 * @brief I2C support library
 *
 * @copyright (C) 2014 Lodz University of Technology, Lodz, Poland  (TUL)
 * @copyright (C) 2014 Deutsches Elektronen-Synchrotron, Hamburg, Germany (DESY)
 * @author Pawel Plewinski (TUL)
 * @author Piotr Perek (TUL)
 * @author Dariusz Makowski (TUL)
 * @par License
 *
 * Redistribution and use in source and binary forms, with or without
 * modification is not allowed. Exceptions are regulated by separate License document. In all cases it is requested to provide the above header.
 *
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#include <asf.h>
#include "i2c_lib.h"
#include <util/delay.h>

int i2c_init(TWI_t *twi)
{
	// send some number of STOP conditions!!
	
	
	twi_options_t opt = {
		.speed = I2C_BAUDRATE,
	};
	
	return twi_master_setup(twi, &opt);
}

int i2c_read(TWI_t *twi, uint8_t i2c_addr, uint8_t* reg_addr, uint8_t reg_addr_length, uint8_t* data_buffer, uint8_t data_length)
{
	int i = 0;
	uint8_t replies = REPLIES_TO_FAIL;
	
	twi_package_t packet_read;
	packet_read.chip         = i2c_addr;					// TWI slave bus address
	if(reg_addr_length > 3)
		return ERR_INVALID_ARG;
	for(i = 0; i < reg_addr_length; i++)
		packet_read.addr[i] = reg_addr[i];					// TWI slave register address
	packet_read.addr_length  = reg_addr_length;				// TWI slave register address size
	packet_read.buffer       = (void *)data_buffer;			// transfer data source buffer
	packet_read.length       = data_length;					// transfer data size (bytes)
	
	while(replies--)
	{
		if(twi_master_read(twi, &packet_read) == TWI_SUCCESS)
			return STATUS_OK;
	}
	
	return ERR_TIMEOUT;
}

int i2c_write(TWI_t *twi, uint8_t i2c_addr, uint8_t* reg_addr, uint8_t reg_addr_length, uint8_t* data_buffer, uint8_t data_length)
{
	int i = 0;
	uint8_t replies = REPLIES_TO_FAIL;

	if(data_length > 32) return ERR_INVALID_ARG;
	
	twi_package_t packet_write;
	packet_write.chip         = i2c_addr;						// TWI slave bus address
	if(reg_addr_length > 3)
	return ERR_INVALID_ARG;
	for(i = 0; i < reg_addr_length; i++)
		packet_write.addr[i] = reg_addr[i];						// TWI slave register address
	packet_write.addr_length  = reg_addr_length;				// TWI slave register address size
	packet_write.buffer       = (void *)data_buffer;			// transfer data source buffer
	packet_write.length       = data_length;					// transfer data size (bytes)
	
	while(replies--)
	{
		if(twi_master_write(twi, &packet_write) == TWI_SUCCESS)
			return STATUS_OK;
	}
	
	return ERR_TIMEOUT;
}


/**
 * \brief Read voltages from power monitor
 * 
 * \param twi pointer to i2c 
 * \param i2c_addr i2c address
 * \param voltage voltage in mV
 * 
 * \return int status
 */
int i2c_power_monitor_read_voltage(TWI_t *twi, uint8_t i2c_addr, uint32_t *voltage)
{
	uint8_t reg_address = REG_VIN_MSB;
	uint8_t data[2] = {0, 0};
	
	if(i2c_read(twi, i2c_addr, &reg_address, 1, data, 2) != TWI_SUCCESS)
		return ERR_IO_ERROR;
	
	*voltage = (((uint32_t)data[0] << 4) | ((uint32_t)data[1] >> 4)) * 25;
	
	return STATUS_OK;
}


/**
 * \brief Read power from power monitor
 *
 * \param twi pointer to i2c
 * \param i2c_addr i2c address
 * \param power power in mW
 * 
 * \return int status
 */
int i2c_power_monitor_read_power(TWI_t *twi, uint8_t i2c_addr, uint32_t *power)
{
	uint8_t reg_address = REG_POWER_MSB2;
	uint8_t data[3] = {0, 0, 0};
	
	if(i2c_read(twi, i2c_addr, &reg_address, 1, data, 3) != TWI_SUCCESS)
		return ERR_IO_ERROR;
	
	*power = (((uint32_t)data[0] << 16) | ((uint32_t)data[1] << 8) | ((uint32_t)data[2])) / 8;	// (25 * 10e-3 * Vin * 25 * Vsense * 10-e6) / 0.005 = 125 * Vin * Vsense * 10e-6 [W] = (Vin * Vsense)/8 [mW]
	
	return STATUS_OK;
}

/**
 * \brief Read current from power monitor
 *
 * \param twi pointer to i2c
 * \param i2c_addr i2c address
 * \param current current in mA
 * 
 * \return int status
 */
int i2c_power_monitor_read_current(TWI_t *twi, uint8_t i2c_addr, uint32_t *current)
{
	uint8_t reg_address = REG_DELTA_SENSE_MSB;
	uint8_t data[2] = {0, 0};
	
	if(i2c_read(twi, i2c_addr, &reg_address, 1, data, 2) != TWI_SUCCESS)
		return ERR_IO_ERROR;
	
	uint32_t dsense = (((uint32_t)data[0] << 4) | ((uint32_t)data[1] >> 4)) * 25;	//in uV
	*current = dsense / 5; //dsense[uV]/5 mOhm = current [mA]
	
	return STATUS_OK;
}

void i2c_power_monitor_wait_for_snapshot(  RelayChannel * channel )
{
	uint8_t flags = 0;
	do
	{
		i2c_read(&LOCAL_I2C, channel->i2c_addr, &reg_control_addr, 1, &flags, 1);
	} while (flags & PM_BUSY_FLAG);
}

bool i2c_power_monitor_start_continuous_measurement( RelayChannel * channel )
{
	return i2c_write(&LOCAL_I2C, channel->i2c_addr, &reg_control_addr, 1, &reg_control_data_continuous, 1) == TWI_SUCCESS;
}

bool i2c_power_monitor_start_snapshot_measurement( RelayChannel * channel )
{
	return i2c_write(&LOCAL_I2C, channel->i2c_addr, &reg_control_addr, 1, &reg_control_data_snapshot, 1) == TWI_SUCCESS;
}

bool i2c_read_uuid(uint8_t * val)
{
	uint8_t addr = I2C_UUID_START_ADDR;
	return i2c_read(&LOCAL_I2C, I2C_UUID_ADDR, &addr, 1, val, I2C_UUID_LEN) == TWI_SUCCESS;
}