/**
 * @file i2c_lib.h
 * @brief I2C support library header
 *
 * @copyright (C) 2014 Lodz University of Technology, Lodz, Poland  (TUL)
 * @copyright (C) 2014 Deutsches Elektronen-Synchrotron, Hamburg, Germany (DESY)
 * @author Pawel Plewinski (TUL)
 * @author Piotr Perek (TUL)
 * @author Dariusz Makowski (TUL)
 * @par License
 *
 * Redistribution and use in source and binary forms, with or without
 * modification is not allowed. Exceptions are regulated by separate License document. In all cases it is requested to provide the above header.
 *
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef I2C_LIB_H_
#define I2C_LIB_H_

#include <asf.h>
#include <stdint.h>
#include <stdbool.h>

#include "power_monitor.h"

#define I2C_BAUDRATE	100000
#define REPLIES_TO_FAIL	10

#define LOCAL_I2C		TWIE

// Addresses of I2C devices on AMC
#define I2C_PWR_MON_P15VD_ADDR				0x6A//0xD4
#define I2C_PWR_MON_P15VA_ADDR				0x6E//0xDC
#define I2C_PWR_MON_P15VH_ADDR				0x67//0xCE
#define I2C_PWR_MON_P7VA_ADDR				0x68//0xD0

#define I2C_PWR_MON_N15VD_ADDR				0x6D//0xDA
#define I2C_PWR_MON_N15VA_ADDR				0x6C//0xD8
#define I2C_PWR_MON_N15VH_ADDR				0x6B//0xD6
#define I2C_PWR_MON_N7VA_ADDR				0x69//0xD2

///Address of UUID EEPROM
#define I2C_UUID_ADDR						0x51//0b0101001
#define I2C_UUID_START_ADDR					0xFA
#define I2C_UUID_LEN						6

#define REG_CONTROL							0x00
#define REG_ALERT							0x01
#define REG_i2c_STATUS						0x02
#define REG_FAULT							0x03
#define REG_FAULT_CoR						0x04
#define REG_POWER_MSB2						0x05
#define REG_POWER_MSB1						0x06
#define REG_POWER_LSB						0x07
#define REG_MAX_POWER_MSB2					0x08
#define REG_MAX_POWER_MSB1					0x09
#define REG_MAX_POWER_LSB					0x0A
#define REG_MIN_POWER_MSB2					0x0B
#define REG_MIN_POWER_MSB1					0x0C
#define REG_MIN_POWER_LSB					0x0D
#define REG_MAX_POWER_THRESHOLD_MSB2		0x0E
#define REG_MAX_POWER_THRESHOLD_MSB1		0x0F
#define REG_MAX_POWER_THRESHOLD_LSB			0x10
#define REG_MIN_POWER_THRESHOLD_MSB2		0x11
#define REG_MIN_POWER_THRESHOLD_MSB1		0x12
#define REG_MIN_POWER_THRESHOLD_LSB			0x13
#define REG_DELTA_SENSE_MSB					0x14
#define REG_DELTA_SENSE_LSB					0x15
#define REG_MAX_DELTA_SENSE_MSB				0x16
#define REG_MAX_DELTA_SENSE_LSB				0x17
#define REG_MIN_DELTA_SENSE_MSB				0x18
#define REG_MIN_DELTA_SENSE_LSB				0x19
#define REG_MAX_DELTA_SENSE_THRESHOLD_MSB	0x1A
#define REG_MAX_DELTA_SENSE_THRESHOLD_LSB	0x1B
#define REG_MIN_DELTA_SENSE_THRESHOLD_MSB	0x1C
#define REG_MIN_DELTA_SENSE_THRESHOLD_LSB	0x1D
#define REG_VIN_MSB							0x1E
#define REG_VIN_LSB							0x1F
#define REG_MAX_VIN_MSB						0x20
#define REG_MAX_VIN_LSB						0x21
#define REG_MIN_VIN_MSB						0x22
#define REG_MIN_VIN_LSB						0x23
#define REG_MAX_VIN_THRESHOLD_MSB			0x24
#define REG_MAX_VIN_THRESHOLD_LSB			0x25
#define REG_MIN_VIN_THRESHOLD_MSB			0x26
#define REG_MIN_VIN_THRESHOLD_LSB			0x27
#define REG_ADIN_MSB						0x28
#define REG_ADIN_LSB						0x29
#define REG_MAX_ADIN_MSB					0x2A
#define REG_MAX_ADIN_LSB					0x2B
#define REG_MIN_ADIN_MSB					0x2C
#define REG_MIN_ADIN_LSB					0x2D
#define REG_MAX_ADIN_THRESHOLD_MSB			0x2E
#define REG_MAX_ADIN_THRESHOLD_LSB			0x2F
#define REG_MIN_ADIN_THRESHOLD_MSB			0x30
#define REG_MIN_ADIN_THRESHOLD_LSB			0x31

int i2c_init(TWI_t *twi);
int i2c_read(TWI_t *twi, uint8_t i2c_addr, uint8_t* reg_addr, uint8_t reg_addr_length, uint8_t* data_buffer, uint8_t data_length);
int i2c_write(TWI_t *twi, uint8_t i2c_addr, uint8_t* reg_addr, uint8_t reg_addr_length, uint8_t* data_buffer, uint8_t data_length);

int i2c_power_monitor_read_voltage(TWI_t *twi, uint8_t i2c_addr, uint32_t *voltage);
int i2c_power_monitor_read_current(TWI_t *twi, uint8_t i2c_addr, uint32_t *current);
int i2c_power_monitor_read_power(TWI_t *twi, uint8_t i2c_addr, uint32_t *power);

bool i2c_power_monitor_start_continuous_measurement( RelayChannel * channel );
bool i2c_power_monitor_start_snapshot_measurement( RelayChannel * channel );
void i2c_power_monitor_wait_for_snapshot(  RelayChannel * channel );

bool i2c_read_uuid(uint8_t * val);
#endif /* I2C_LIB_H_ */