﻿/**
 * @file mgmt_protocol_master.c
 * @brief Management protocol between TMCB2 and FRED boards support library 
 *
 * @copyright (C) 2014 Lodz University of Technology, Lodz, Poland  (TUL)
 * @copyright (C) 2014 Deutsches Elektronen-Synchrotron, Hamburg, Germany (DESY)
 * @author Pawel Plewinski (TUL)
 * @author Dariusz Makowski (TUL)
 * @par License
 *
 * Redistribution and use in source and binary forms, with or without
 * modification is not allowed. Exceptions are regulated by separate License document. In all cases it is requested to provide the above header.
 *
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "i2c_lib.h"
#include "twim.h"
#include "mgmt_protocol_master.h"
#include "usart_lib.h"
#include "telnet_protocol.h"
#include "power_monitor.h"
#include "cpld_protocol.h"

#include <stdio.h>

typedef enum {Not_configured, Disabling, Configured} FpgaConfigureState;
FpgaConfigureState tmcb_fpga_state = Not_configured;
uint16_t tmcb_status = 0, tmcb_prev_status;

#define CHANGED(val, prev, byte) ((val ^ prev) & byte))

/**
 * \brief Initialize management i2c protocol 
 * 
 * 
 * \return void
 */
void mgmtInit( void )
{
	i2c_init(&MANAGE_I2C);
}

/**
 * \brief Read register on mgmt bus
 * 
 * \param reg Register
 * \param data Pointer to data
 * 
 * \return int Status
 */
int mgmtReadRegister( uint8_t reg, uint16_t *data)
{
	if ( i2c_read(&MANAGE_I2C, MANAGE_ADDR,  &reg, 1, (uint8_t*) data, 2) == TWI_SUCCESS)
	{
		setCpldErrorLED(false);
		return STATUS_OK;
	}
	else
	{
		setCpldErrorLED(true);
		return ERR_IO_ERROR;
	}
}

int mgmtWriteRegister( uint8_t reg, uint16_t data )
{
	uint8_t replies = REPLIES_TO_FAIL;
	twi_package_t packet_write;
	packet_write.addr_length  = 1;					// TWI slave memory address data size
	packet_write.addr[0]	  = reg;
	packet_write.chip         = MANAGE_ADDR;		// TWI slave bus address
	packet_write.buffer       = (void *)&data;		// transfer data source buffer
	packet_write.length       = 2;					// transfer data size (bytes)
	
	while(replies--)
	{
		if(twi_master_write(&MANAGE_I2C, &packet_write) == TWI_SUCCESS)
		{
			setCpldErrorLED(false);
			return STATUS_OK;
			
		}
	}
	setCpldErrorLED(true);	
	return ERR_TIMEOUT;
}

/**
 * \brief Read voltage from TMCB2
 * 
 * \param sensor
 * 
 * \return uint16_t
 */
uint16_t mgmtReadVoltage( uint8_t sensor )
{
	uint16_t result = 0;
	mgmtReadRegister(REG_VOLTAGE_FIRST | (sensor & 0x0F), &result);
	return result;
}


/**
 * \brief Handling hot-plug management of FRED board
 * 
 * 
 * \return void
 */
void pollMgmt(void)
{
	tmcb_status = 0;
	if (getTmcbStatus(&tmcb_status) == STATUS_OK)
	{
		uint16_t channelMask = 0x00FF;
		bool change = false;
		///TMCB_HANDLE_MODE configurable in conf_board.h
		#if (TMCB_HANDLE_MODE == TMCB_HANDLE_MODE_0)
		if (mgmtReadRegister(REG_FRED_CHANNEL_MASK, &channelMask) != STATUS_OK)
			channelMask = VOLTAGE_MASK_P15VD | VOLTAGE_MASK_P15VA | VOLTAGE_MASK_P15VH
			| VOLTAGE_MASK_P7VA | VOLTAGE_MASK_N15VD | VOLTAGE_MASK_N15VA
			| VOLTAGE_MASK_N7VA | VOLTAGE_MASK_N15VH;
		#endif
		#if (TMCB_HANDLE_MODE == TMCB_HANDLE_MODE_1 || TMCB_HANDLE_MODE == TMCB_HANDLE_MODE_3)
			channelMask = VOLTAGE_MASK_P15VD | VOLTAGE_MASK_N15VD;
		#endif
		#if (TMCB_HANDLE_MODE == TMCB_HANDLE_MODE_2)
			channelMask = VOLTAGE_MASK_P15VD | VOLTAGE_MASK_P15VA | VOLTAGE_MASK_P15VH
			| VOLTAGE_MASK_P7VA | VOLTAGE_MASK_N15VD | VOLTAGE_MASK_N15VA
			| VOLTAGE_MASK_N7VA| VOLTAGE_MASK_N15VH;
		#endif
		
		
		if (tmcb_fpga_state == Not_configured && (tmcb_status & REG_STATUS_AMC_HANDLE))
		{
			setTmcbBlueLed(REG_BLUE_LED_BLINK);
			if ((tmcb_status & REG_STATUS_ALLOW_ENABLE) != 0)
			{
				//Turn on voltages
				bool allow_on = true;
				#if (TMCB_HANDLE_MODE == TMCB_HANDLE_MODE_3)
					allow_on = false;
					for (int i = 0; i < VOLTAGE_COUNT; ++i)
						if (channels[i].actual_state == true)
							allow_on = true;
				#endif
				for (int i = 0; i < VOLTAGE_COUNT; ++i)
				{
					if (channelMask & (1 << i) && channels[i].state == false && allow_on == true)
					{
						channels[i].state = 1;
						channels[i].state_changed = true;
						change = true;
					}
				}
				mgmtWriteRegister(REG_ENABLE_FPGA, REG_FPGA_ENABLE);
				setTmcbBlueLed(REG_BLUE_LED_OFF);
				tmcb_fpga_state = Configured;

			}
		}
		else if ((tmcb_fpga_state != Not_configured && tmcb_fpga_state != Disabling) && ((tmcb_status & REG_STATUS_AMC_HANDLE) == 0))
		{
			if ((tmcb_status & REG_STATUS_ALLOW_DISABLE) != 0)
			{
				mgmtWriteRegister(REG_ENABLE_FPGA, REG_FPGA_DISABLE);
				tmcb_fpga_state = Disabling;
				setTmcbBlueLed(REG_BLUE_LED_BLINK);
			}
		}
		if (tmcb_fpga_state == Disabling && ((tmcb_status & REG_STATUS_FPGA_CONFIGURED) == 0))
		{
			if ((tmcb_status & REG_STATUS_ALLOW_DISABLE) != 0)
			{
				//Turn off voltages
				for (int i = 0; i < VOLTAGE_COUNT; ++i)
				{
					if (channelMask & (1 << i))// && channels[i].state != false)
					{
						channels[i].state = 0;
						channels[i].state_changed = true;
						change = true;
					}
				}
			}
			setTmcbBlueLed(REG_BLUE_LED_ON);
			tmcb_fpga_state = Not_configured;
		}
		if (change){
			pollChannels(false);
			writeRelays();
		}
	}
}


int getTmcbStatus( uint16_t *status )
{
	int res = mgmtReadRegister(REG_TMCB_STATUS, status);
	return res;
}

int setTmcbBlueLed( uint16_t value )
{
	return mgmtWriteRegister(REG_BLUE_LED, value);
}

bool canPowerUp( void )
{
	return (tmcb_status & REG_STATUS_ALLOW_ENABLE);
}

bool canPowerDown( void )
{
	return (tmcb_status & REG_STATUS_ALLOW_DISABLE);
}


/**
 * \brief Test management protocol and print results on serial console
 * 
 * 
 * \return void
 */
void testMgmtProtocol(void)
{
	char msg[80];
	const char *names[] = {
		"REG_TEST", "REG_POWER_GOOD", "REG_FPGA_CONTROL",
		"REG_FPGA_PROM", "REG_TEMPERATURE", "REG_FMC", "REG_SW",
		"REG_TEMPERATURE_STATUS"
	};
	const char* adc_channel_name[] = { "+15VD",
		"+12VA",
		"+10VA",
		"+5VA",
		"+5VD",
		"+3V3D",
		"+2V5A",
		"-5VA",
		"-10VA",
		"-12VA",
		"-15VD"
	};
	uint16_t data;
	printPString("*****************\r\n ");
	for (int i = 0; i < sizeof(names)/sizeof(names[0]); ++i)

	{
		mgmtReadRegister(i, &data);
		sprintf_P(msg, PROGMEM_STRING("Register %s value %X\r\n"), names[i], data);
		printString(msg);
	}
	for (int i = 0; i < sizeof(adc_channel_name)/sizeof(adc_channel_name[0]); ++i)

	{
		sprintf_P(msg, PROGMEM_STRING("Voltage %s value %X\r\n"), adc_channel_name[i], mgmtReadVoltage(i));
		printString(msg);		
	}
}

bool tmcbIsConfigured( void )
{
	return tmcb_fpga_state == Configured;
}
