﻿/**
 * @file power_monitor.c
 * @brief Library for communication with power monitors and handling channel info
 *
 * @copyright (C) 2014 Lodz University of Technology, Lodz, Poland  (TUL)
 * @copyright (C) 2014 Deutsches Elektronen-Synchrotron, Hamburg, Germany (DESY)
 * @author Pawel Plewinski (TUL)
 * @author Piotr Perek (TUL)
 * @author Dariusz Makowski (TUL)
 * @par License
 *
 * Redistribution and use in source and binary forms, with or without
 * modification is not allowed. Exceptions are regulated by separate License document. In all cases it is requested to provide the above header.
 *
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "power_monitor.h"
#include "i2c_lib.h"
#include "cpld_protocol.h"
#include "mgmt_protocol_master.h"
#include "eeprom.h"
#include "version.h"
#include <avr/eeprom.h>
#include <util/delay.h>


RelayChannel channels[8];


uint8_t reg_control_addr = REG_CONTROL;
uint8_t reg_control_data_continuous = 0x01;
uint8_t reg_control_data_snapshot = 0x81;


/**
 * \brief Initialize all channels
 * 
 * \param 
 * 
 * \return void
 */
void initChannelStructures( void )
{
	//P15VD
	channels[0].i2c_addr = I2C_PWR_MON_P15VD_ADDR;
	channels[0].negative = false;
	channels[0].eeprom_addr = EEPROM_ADDR_LIMIT_CHANNEL0;
	//P15VA
	channels[1].i2c_addr = I2C_PWR_MON_P15VA_ADDR;
	channels[1].negative = false;
	channels[1].eeprom_addr = EEPROM_ADDR_LIMIT_CHANNEL1;
	//P15VH
	channels[2].i2c_addr = I2C_PWR_MON_P15VH_ADDR;
	channels[2].negative = false;
	channels[2].eeprom_addr = EEPROM_ADDR_LIMIT_CHANNEL2;
	//P7VA
	channels[3].i2c_addr = I2C_PWR_MON_P7VA_ADDR;
	channels[3].negative = false;
	channels[3].eeprom_addr = EEPROM_ADDR_LIMIT_CHANNEL3;
	//N15VD
	channels[4].i2c_addr = I2C_PWR_MON_N15VD_ADDR;
	channels[4].negative = true;
	channels[4].eeprom_addr = EEPROM_ADDR_LIMIT_CHANNEL4;
	//N15VA
	channels[5].i2c_addr = I2C_PWR_MON_N15VA_ADDR;
	channels[5].negative = true;
	channels[5].eeprom_addr = EEPROM_ADDR_LIMIT_CHANNEL5;
	//N7VA
	channels[6].i2c_addr = I2C_PWR_MON_N7VA_ADDR;
	channels[6].negative = true;
	channels[6].eeprom_addr = EEPROM_ADDR_LIMIT_CHANNEL6;
	//N15VH
	channels[7].i2c_addr = I2C_PWR_MON_N15VH_ADDR;
	channels[7].negative = true;
	channels[7].eeprom_addr = EEPROM_ADDR_LIMIT_CHANNEL7;
	
	readRelays();

	for (int i = 0; i < CHANNEL_NUMBER; ++i)
	{
		channels[i].alarm_mask = 
			(eeprom_read_byte(EEPROM_ADDR_LIMIT_MASK) & (1 << i)) ? true : false;
		channels[i].overcurrent_alarm_status = 0;
		channels[i].overcurrent_alarm_counter = 0;
		channels[i].alarm_pin0_status = false;
		channels[i].alarm_pin1_status = false;
		channels[i].current_limit = eeprom_read_dword(channels[i].eeprom_addr);
		channels[i].state = channels[i].actual_state;
		channels[i].i2c_error = false;
		channels[i].state_changed = true;
	}
	
	//Fred 3m - only positive channels
	if (get_version() == FRED3M)
	{
		for (int i = 0; i < CHANNEL_NUMBER; ++i)
			channels[i].negative = false;
	}

	
	//enable portC interrupts
	PORTC.INT0MASK = ioport_pin_to_mask(SWITCH_UP) | ioport_pin_to_mask(SWITCH_DOWN);
	PORTC.INT1MASK = ioport_pin_to_mask(CPLD_INTERRUPT);
	PORTC.INTCTRL = (PORT_INT0LVL_HI_gc | PORT_INT1LVL_HI_gc);
}


///ISR for switch
ISR(PORTC_INT0_vect)
{
	if (ioport_pin_is_high(SWITCH_UP))
	{
		//printString("Switch up\r\n");
		for (int i = 0; i < CHANNEL_NUMBER; ++i)
		{
			channels[i].state = true;
			channels[i].state_changed = true;
		}
	}
	else if (ioport_pin_is_high(SWITCH_DOWN))
	{
		//printString("Switch down\r\n");
		for (int i = 0; i < CHANNEL_NUMBER; ++i)
		{
			channels[i].state = false;
			channels[i].state_changed = true;
		}
	}
}

///ISR for CPLD interrupt
ISR(PORTC_INT1_vect)
{
	if (ioport_pin_is_high(CPLD_INTERRUPT))
	{
		readRelays();
	}
}


/**
 * \brief Update channel
 * 
 * \param channel Channel
 * \param nr Channel number
 * \param timerTriggerred Flag set if the function is invoked by timer event, not by commands
 * 
 * \return void
 */
void pollChannel( RelayChannel *channel, int nr, bool timerTrigerred )
{
	if(i2c_write(&LOCAL_I2C, channel->i2c_addr, &reg_control_addr, 1, &reg_control_data_continuous, 1) == TWI_SUCCESS)
	{
		i2c_power_monitor_read_voltage(&LOCAL_I2C, channel->i2c_addr, &(channel->voltage));
		i2c_power_monitor_read_current(&LOCAL_I2C, channel->i2c_addr, &(channel->current));
		i2c_power_monitor_read_power(&LOCAL_I2C, channel->i2c_addr, &(channel->power));
		channel->i2c_error = false;
	}
	else
	{
		channel->voltage = 0;
		channel->current = 0;
		channel->i2c_error = true;
	}
	if (channel->overcurrent_alarm_status & channel->actual_state)
	{
		channel->overcurrent_alarm_status = 0;
	}
	if ((channel->current >= channel->current_limit) && channel->alarm_mask)
	{
		if (timerTrigerred)
		{
			channel->overcurrent_alarm_counter++;
		}
		if (channel->overcurrent_alarm_counter > OVERCURRENT_LIMIT_COUNTER)
		{	
			channel->state_changed = true;
			channel->overcurrent_alarm_status = true;
		}
	}
	else
	{
		channel->overcurrent_alarm_counter = 0;
	}
	if ((readRegister(CPLD_ADDR_ALARMPIN_TRIGGERED) & 0x1) 
		&& ((readRegister(CPLD_ADDR_ALARMPIN0_MASK)) & (1 << nr)))
	{
		channel->alarm_pin0_status = true;
		channel->state = 0;
		channel->state_changed = true;
	}
	else
	{
		channel->alarm_pin0_status = false;
	}
	if ((readRegister(CPLD_ADDR_ALARMPIN_TRIGGERED) & 0x2)
		&& ((readRegister(CPLD_ADDR_ALARMPIN1_MASK)) & (1 << nr)))
	{
		channel->alarm_pin1_status = true;
		channel->state = 0;
		channel->state_changed = true;
	}
	else
	{
		channel->alarm_pin1_status = false;
	}
	
}

/**
 * \brief Update all channels
 * 
 * 
 * \return void
 */
void pollChannels(bool timerTriggerred)
{
	for (int i = 0; i < CHANNEL_NUMBER; ++i)
	{
		pollChannel(&channels[i], i, timerTriggerred);
	}
	readRelays();
}

uint8_t getAlarmMask()
{
	uint8_t res = 0;
	for (int i = 0; i < CHANNEL_NUMBER; ++i)
	{
		res |= ((channels[i].alarm_mask?1:0) << i);
	}
	return res;
}

void setAlarmMask(uint8_t alarmMask)
{
	for (int i = 0; i < CHANNEL_NUMBER; ++i)
	{
		if ((alarmMask >> i) & 0x01)
		channels[i].alarm_mask = true;
		else
		channels[i].alarm_mask = false;
	}
	eeprom_write_byte(EEPROM_ADDR_LIMIT_MASK, alarmMask);
}

/**
 * \brief Set current limit for a channel, not greater than MAX_LIMIT
 * 
 * \param channel Channel number
 * \param limit Limit
 * 
 * \return void
 */
void setCurrentLimit(int channel, uint32_t limit)
{
	if (limit < MAX_CURRENT_LIMIT)
	channels[channel].current_limit = limit;
	else
	channels[channel].current_limit = MAX_CURRENT_LIMIT;
	eeprom_write_dword(channels[channel].eeprom_addr, channels[channel].current_limit);
}

/**
 * \brief Read state of relays
 * 
 * \param 
 * 
 * \return void
 */
void readRelays( void )
{
	int reg_val = getRelaysState();
	for (int i = 0; i < CHANNEL_NUMBER; ++i)
	{
		channels[i].actual_state = (reg_val & (1 << i))? true : false;
	}
}

/**
 * \brief Set state on relays
 * 
 * \param 
 * 
 * \return void
 */
void writeRelays( void )
{
	int reg_val = readRegister(CPLD_ADDR_RELAY_STATUS);
	bool changed = false;
	for ( int i = 0; i < CHANNEL_NUMBER; ++i)
	{
		if (channels[i].state_changed )
		{
			if (channels[i].state & !(channels[i].overcurrent_alarm_status))
			{
				uint16_t val = 0;
				//Check if TMCB allows to switch power on
				if (ioport_pin_is_high(CPLD_TMCB_JUMPER))
				{
					if (canPowerUp())
					{
							reg_val |= (1 << i);
							changed = true;
					}
				}
				//If TMCB jumper not present, switch on
				else 
				{
					if (ioport_pin_is_high(CPLD_TMCB_JUMPER))
					{
						if (canPowerDown())
						{
							reg_val |= (1 << i);
							changed = true;
						}
					}
					else
					{
						reg_val |= (1 << i);
						changed = true;
					}

				}
			}
			else
			{
				reg_val &= ~(1 << i);
				changed = true;
			}
			if (channels[i].overcurrent_alarm_status)
			{
				reg_val &= ~(1 << i);
				changed = true;
			}
		}
		channels[i].state_changed = false;
	}
	if (changed)
	{
		setRelaysState(reg_val);
	}
	_delay_us(50);
}

uint8_t getPowerState()
{
	return readRegister(CPLD_ADDR_RELAY_CONTROL);
}

void clearOvercurrentAlarm( int channel )
{
/*	if (channels[channel].state & channels[channel].overcurrent_alarm_status)
	{
		channels[channel].state = false;
	}*/
	channels[channel].overcurrent_alarm_status = false;
}


