﻿/**
 * @file telnet_protocol.c
 * @brief Telnet protocol for Xport
 *
 * @copyright (C) 2014 Lodz University of Technology, Lodz, Poland  (TUL)
 * @copyright (C) 2014 Deutsches Elektronen-Synchrotron, Hamburg, Germany (DESY)
 * @author Pawel Plewinski (TUL)
 * @author Dariusz Makowski (TUL)
 * @par License
 *
 * Redistribution and use in source and binary forms, with or without
 * modification is not allowed. Exceptions are regulated by separate License document. In all cases it is requested to provide the above header.
 *
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "telnet_protocol.h"
#include "power_monitor.h"
#include "usart_lib.h"
#include "cpld_protocol.h"
#include "user_board.h"
#include "i2c_lib.h"
#include "mgmt_protocol_master.h"
#include "fan_lib.h" // added by Stanislav Chystiakov
#include "fan_current_monitor_lib.h" // added by Stanislav Chystiakov
#include "eeprom.h"
#include <util/delay.h>
#include <usart_serial.h>
#include <stdbool.h>
#include <string.h>
#include <ctype.h>
#include <stdio.h>
#include <math.h>

volatile bool bufferReady = false;
volatile int inputBufferCounter = 0;
volatile char inputBuffer[INPUT_BUFFER_SIZE] = {0};




void clearBuffer( void )
{
	inputBuffer[0] = '\0';
	inputBufferCounter = 0;
	bufferReady = false;
	usart_set_rx_interrupt_level(USART_SERIAL, USART_INT_LVL_LO);
	(USART_SERIAL)->CTRLA |= (0x1 << 4);
}

void telnetPutChar(int ch)
{
	if (ch == '\b' && inputBufferCounter > 0)
	{
		inputBufferCounter--;
	}
	if (ch == '\r')
	{
		inputBuffer[inputBufferCounter] = '\0';
		bufferReady = true;
		usart_set_rx_interrupt_level(USART_SERIAL, USART_INT_LVL_OFF);
		(USART_SERIAL)->CTRLA &=  ~(0x11 << 4);
	}
	if (!isprint(ch))
	{
		return;
	}
	if (inputBufferCounter < INPUT_BUFFER_SIZE)
	inputBuffer[inputBufferCounter++] = tolower(ch);
	else
	{
		inputBuffer[INPUT_BUFFER_SIZE-1]='\0';
		usart_set_rx_interrupt_level(USART_SERIAL, USART_INT_LVL_OFF);
		(USART_SERIAL)->CTRLA &= ~(0x11 << 4);
		bufferReady = true;
	}
}

void channelsOn()
{
	for (int i = 0; i < CHANNEL_NUMBER; ++i)
		{
			channels[i].state = 1;
			channels[i].state_changed = true;
		}
	pollChannels(false);
	writeRelays();
}

void channelsOff()
{
	for (int i = 0; i < CHANNEL_NUMBER; ++i)
		{
			channels[i].state = 0;
			channels[i].state_changed = true;
		}
	pollChannels(false);
	writeRelays();
}


/**
 * \brief 
 * Parses the command from inputBuffer
 * \param 
 * 
 * \return void
 */
void parseCommand( void )
{
	///Supported commands:
	if (strstr( (char*) inputBuffer, "dci") == inputBuffer
	 || strstr( (char*) inputBuffer, "ci") == inputBuffer)
	{
		//WELCOME MESSAGE
		welcomeMessage();
		printVersion();
		printPrompt();
	}
	else if (strstr( (char*) inputBuffer, "readchan") == inputBuffer)
	{
		///READ CHANNEL
		unsigned int channel;
		if (sscanf( (char*) inputBuffer, "readchan %u", &channel) != 1 || channel > 7)
		{
			printBadArgument();
		}
		else
		{
			channelsSnapshot();
			printChannelHeader();
			printChannel(channel);
		}
	}
	else if (strstr( (char*) inputBuffer, "readall") == inputBuffer)
	{
		///READ ALL CHANNELS
		channelsSnapshot();
		printChannelHeader();
		for (int i = 0; i < CHANNEL_NUMBER; ++i)
		{
			printChannel(i);
		}
	}
	else if (strstr( (char*) inputBuffer, "setpwrstates") == inputBuffer)
	{
		///SET POWER STATES
		unsigned powerStateWord, channelMask;
		if (sscanf( (char*) inputBuffer, "setpwrstates %2X %2X", &powerStateWord, &channelMask) != 2)
		{
			printBadArgument();
		}
		else
		{
			for (int i = 0; i < CHANNEL_NUMBER; ++i)
			{
				if ( (channelMask >> i) & 0x01 )
				{
					if ( (powerStateWord >> i) & 0x01)
					{
						channels[i].state = true;
						channels[i].state_changed = true;
					}
					else
					{
						channels[i].state = false;
						channels[i].state_changed = true;							
					}
				}
			}
			pollChannels(false);
			writeRelays();
			printPowerStateWord();
		}
	}
	else if (strstr( (char*) inputBuffer, "chanon") == inputBuffer)
	{
		///TURN CHANNEL ON
		unsigned int channel;
		if (sscanf( (char*) inputBuffer, "chanon %u", &channel) != 1 || channel > 7)
		{
			printBadArgument();
		}
		else
		{
			channels[channel].state = 1;
			channels[channel].state_changed = true;
			writeRelays();
			printPowerStateWord();
		}
	}
	else if (strstr( (char*) inputBuffer, "chanoff") == inputBuffer)
	{
		///TURN CHANNEL OFF
		unsigned int channel;
		if (sscanf( (char*) inputBuffer, "chanoff %u", &channel) != 1 || channel > 7)
		{
			printBadArgument();
		}
		else
		{
			channels[channel].state = 0;
			channels[channel].state_changed = true;
			writeRelays();
			printPowerStateWord();
		}
	}
	
	else if (strstr( (char*) inputBuffer, "readcurlimits") == inputBuffer)
	{
		///READ ALL CHANNELS' CURRENT LIMITS
		printLimits();
	}
	else if (strstr( (char*) inputBuffer, "progcurlimit") == inputBuffer)
	{
		///SET CURRENT LIMIT
		uint32_t channel, limit;
		if (sscanf( (char*) inputBuffer, "progcurlimit %"PRIu32" %"PRIu32"", &channel, &limit) != 2 || channel > 7)
		{
			printBadArgument();
		}
		else
		{
			setCurrentLimit(channel,limit);
		//	clearOvercurrentAlarm(channel);
			char li[16];
			sprintf(li, "%"PRIu32"", channels[channel].current_limit);
			printString(li);
		}
	}
	else if (strstr( (char*) inputBuffer, "poweron") == inputBuffer)
	{
		///TURN ALL CHANNELS ON
		channelsOn();
		printPowerStateWord();
	}
	else if (strstr( (char*) inputBuffer, "poweroff") == inputBuffer)
	{
		///TURN ALL CHANNELS OFF
		channelsOff();
		printPowerStateWord();
	}
	else if (strstr( (char*) inputBuffer, "setalarmmasks") == inputBuffer)
	{
		///SET ALARM MASKS
		uint8_t alarmMasks, pin0Mask, pin1Mask;
		if (sscanf( (char*) inputBuffer, "setalarmmasks %2"PRIX8" %2"PRIX8" %2"PRIX8,
			&alarmMasks, & pin0Mask, &pin1Mask) != 3)
		{
			printBadArgument();
		}
		else
		{
			setAlarmMask(alarmMasks);
			printAlarmMasks();
			setAlarmPin0Mask(pin0Mask);
			setAlarmPin1Mask(pin1Mask);
			printPinAlarmMasks();
		}
	}
	else if (strstr( (char*) inputBuffer, "readalarms") == inputBuffer)
	{
		///READ ALARMS
		char buf[4];
		sprintf(buf,"%02X",readRegister(CPLD_ADDR_ALARMPIN_TRIGGERED));
		printString(buf);
	}
	else if (strstr( (char*) inputBuffer, "clearalarms") == inputBuffer)
	{
		///CLEAR ALARMS
		writeRegister(CPLD_ADDR_ALARMPIN_TRIGGERED, 0x00);
		for (int i = 0; i < CHANNEL_NUMBER; ++i)
		{
			channels[i].alarm_pin0_status = false;
			channels[i].alarm_pin1_status = false;
			clearOvercurrentAlarm(i);
		}
		char buf[4];
		sprintf(buf,"%02X",readRegister(CPLD_ADDR_ALARMPIN_TRIGGERED));
		printString(buf);
	}
	else if (strstr( (char*) inputBuffer, "readalarmmasks") == inputBuffer)
	{
		printAlarmMasks();
		printPinAlarmMasks();
	}
	else if (strstr( (char*) inputBuffer, "tmcbselectprom") == inputBuffer)
	{
		///SET FPGA PROM
		uint8_t value;
		if (sscanf( (char*) inputBuffer, "tmcbselectprom %2X", &value) != 1)
		{
			printBadArgument();
		}
		else
		{
			uint16_t data = 0;
			if (mgmtReadRegister(REG_FPGA_CONTROL, &data) != ERR_IO_ERROR)
			{
				if (value)
					data |= REG_FPGA_CONTROL_REV_SEL1;
				else
					data &= (~REG_FPGA_CONTROL_REV_SEL1);
				if (mgmtWriteRegister(REG_FPGA_CONTROL, data) == TWI_SUCCESS)
				{
					printString("Success\r\n");
				}
				else
					printString("Error in communication with TMCB!\r\n");
			}
			else
				printString("Error in reading from TMCB!\r\n");
		}
	}
	else if (strstr( (char*) inputBuffer, "tmcbreadfpgastatus") == inputBuffer)
	{
		///READ FPGA STATUS
		uint16_t data = 0;
		if (mgmtReadRegister(REG_FPGA_CONTROL, &data) != ERR_IO_ERROR)
		{
			char buf[40];
			sprintf(buf,"%02X",(data & REG_FPGA_CONTROL_DONE_N) ? 1 : 0);
			printString(buf);
		}
		else
		printString("Error in reading from TMCB!\r\n");

	}
	else if (strstr( (char*) inputBuffer, "tmcbreloadfpga") == inputBuffer)
	{
		///SET FPGA PROG B
			uint16_t data = 0;
			if (mgmtReadRegister(REG_FPGA_CONTROL, &data) != ERR_IO_ERROR)
			{
				data &= (~REG_FPGA_CONTROL_PROG);
				mgmtWriteRegister(REG_FPGA_CONTROL, data);
				_delay_ms(100);
				data |= REG_FPGA_CONTROL_PROG;
				mgmtWriteRegister(REG_FPGA_CONTROL, data);
			}
			else
				printString("Error in reading from TMCB!\r\n");
	}
	else if (strstr( (char*) inputBuffer, "tmcbprogpinvoltage") == inputBuffer)
	{
		///SET FPGA PROM
		unsigned pin, voltage;
		if (sscanf( (char*) inputBuffer, "tmcbprogpinvoltage %2d %2d", &pin, &voltage) != 2)
		{
			printBadArgument();
		}
		else if (pin > 20 || (voltage != 3 && voltage != 5))
		{
			printString("\r\nSelect pin in range 1-20, voltage 3 or 5");
			printBadArgument();
		}
		else
		{
			uint16_t data = 0;
			if (mgmtReadRegister(REG_SW, &data) != ERR_IO_ERROR)
			{
				char buf[20];
				sprintf(buf, "%X   ", data);
				printString(buf);
				uint8_t pin_no = (pin - 1) / 2; 
				if (voltage == 3)
					data |= (1 << pin_no);
				else
					data &= (~(1 << pin_no));
				sprintf(buf, "%X   ", data);
				printString(buf);				
				if (mgmtWriteRegister(REG_SW, data) == TWI_SUCCESS)
				{
					printString("Success\r\n");
				}
				else
					printString("Error in communication with TMCB!\r\n");
			}
			else
			printString("Error in reading from TMCB!\r\n");
		}
	}	
	else if (strstr( (char*) inputBuffer, "tmcbreadpinvoltages") == inputBuffer)
	{
		uint16_t data = 0;
		if (mgmtReadRegister(REG_SW, &data) != ERR_IO_ERROR)
		{
			for (int i = 1; i < 20; i += 2)
			{
				char buf[48];
				sprintf(buf, "Pins %d-%d - %sV\r\n", i, i+1, 
					(data & (1 << ((i-1) / 2))) ? "3.3" : "5");
				printString(buf);
			}
		}
		else
			printString("Error in communication with TMCB!\r\n");
	}
	else if (strstr( (char*) inputBuffer, "exit") == inputBuffer)
	{
		printString("Command not implemented");
	}
	else if (strstr( (char*) inputBuffer, "version") == inputBuffer)
	{
		printVersion();
	}
	else if (strstr( (char*) inputBuffer, "fwupgrade") == inputBuffer)
	{
		eeprom_write_upgrade();
		//reset CPU
		ccp_write_io((void *)&RST.CTRL, RST_SWRST_bm);
	}
	else if (strstr( (char*) inputBuffer, "progdevicename ") == inputBuffer)
	{
		eeprom_write_name((char *) &(inputBuffer[15]));
	}
	else if (strstr( (char*) inputBuffer, "proginvnr ") == inputBuffer)
	{
		eeprom_write_inv_no((char *) &(inputBuffer[10]));
	}
	
	else if (strstr( (char*) inputBuffer, "a0") == inputBuffer)
	{
		writeRegister(CPLD_ADDR_ALARMPIN_TRIGGERED, 
			readRegister(CPLD_ADDR_ALARMPIN_TRIGGERED) | 0x1);
		char buf[4];
		sprintf(buf,"%02X",readRegister(CPLD_ADDR_ALARMPIN_TRIGGERED));
		printString(buf);
	}
	else if (strstr( (char*) inputBuffer, "a1") == inputBuffer)
	{
		writeRegister(CPLD_ADDR_ALARMPIN_TRIGGERED, 
			readRegister(CPLD_ADDR_ALARMPIN_TRIGGERED) | 0x2);
		char buf[4];
		sprintf(buf,"%02X",readRegister(CPLD_ADDR_ALARMPIN_TRIGGERED));
		printString(buf);
	}
	else if (strstr( (char*) inputBuffer, "help") == inputBuffer)
	{
		///HELP
		printString("readChan chanNr\r\n"
		"readAll\r\n"
		"setPwrStates\r\n"
		"chanOn chanNr\r\n"
		"chanOff chanNr\r\n"
		"readCurLimits\r\n"
		"progCurLimit chanNr currentLimit\r\n"
		"powerOn\r\n"
		"powerOff\r\n"
		"readAlarms\r\n"
		"clearAlarms\r\n"
		"setAlarmMasks overcurrentAlarmMask pin0AlarmMask pin1AlarmMask\r\n"
		"readAlarmMasks\r\n"
		"\r\n"
		"TmcbSelectProm value\r\n"
		"TmcbReloadFPGA\r\n"
		"TmcbReadFPGAstatus\r\n"
		"TmcbProgPinVoltage pin voltage\r\n"
		"TmcbReadPinVoltages\r\n"
		"\r\n"
		"FbProgFan mode speed\r\n"
		"FbReadFansInfo\r\n"
		"FbReadFanStates\r\n"
		"FbSetIOportDir port value\r\n"
		"FbSetIOportLevel port value\r\n"
		"FbReadIOportLevel\r\n"
		"\r\n"
		// --> added by Stanislav Chystiakov
		"help\r\n"
		"progDeviceName name\r\n"
		"progInvNr invNumber\r\n"
		"version\r\n"
		"fwupgrade");
	}
	//=====================================================================================================
	//added by Stanislav Chystiakov
	else if (strstr( (char*) inputBuffer, "fbprogfan") == inputBuffer) /// set FAN speed. Max speed range is 255!
	{	
		unsigned int speedRange;
		char mode;
		if ((sscanf( (char*) inputBuffer, "fbprogfan %c %u", &mode, &speedRange) != 2) || (mode != 'v' && mode != 'r'))
		{
			printString("Supported modes - R, V\r\n");
			printBadArgument();
		}
		else if (mode == 'v' && ((speedRange > 0) || speedRange < 255))
		{
			setFanSpeed(speedRange);
		}
		else if (mode == 'r')
		{
			setFanSpeedRpm(speedRange);
		}
		else
		{
			printString("For Voltage mode - value in range 0-255");
			printBadArgument();
		}
	}
	else if (strstr( (char*) inputBuffer, "fbsetaverage") == inputBuffer) /// set average
	{
		unsigned int avrgVal;
		if ((sscanf( (char*) inputBuffer, "fbsetaverage %u", &avrgVal) != 1) || (avrgVal < 1) || (avrgVal > 255))
		{
			printBadArgument();
		}
		else
		{
			setAverage(avrgVal);
		}
	}
	
	else if (strstr( (char*) inputBuffer, "fbreadfansinfo") == inputBuffer) ///read info for all fans
	{
		printFanInfoHeader();
		for (int i = 0; i < FAN_COUNT; ++i)
			printFanInfo(i);
	}
	else if (strstr( (char*) inputBuffer, "fbsetioportdir" ) == inputBuffer) /// set port direction
	{
		unsigned int portDir;
		char port;
		if ((sscanf( (char*) inputBuffer, "fbsetioportdir %c %x", &port, &portDir) != 2) || (portDir > 255)) // input in HEX
		{
			printBadArgument();
		}
		else if (port == 'b')
		{
			setPortDirection(_PORTB_, portDir);
		}
		else if (port == 'd')
		{
			setPortDirection(_PORTD_, portDir);
		}
		else{
			printString("Supported ports - B and D");
			printBadArgument();
		}
	}
	
	else if (strstr( (char*) inputBuffer, "fbsetioportlevel") == inputBuffer) ///set port level
	{
		unsigned int portLev;
		char port;
		if ((sscanf( (char*) inputBuffer, "fbsetioportlevel %c %x", &port, &portLev) != 2) || (portLev > 255)) // input in HEX
		{
			printBadArgument();
		}
		else if (port == 'b')
		{
			setPortLevel(_PORTB_, portLev);
		}
		else if (port == 'd')
		{
			setPortLevel(_PORTD_, portLev);
		}
		else{
			printString("Supported ports - B and D");
			printBadArgument();
		}
	}
	else if (strstr( (char*) inputBuffer, "fbreadioportlevel") == inputBuffer) ///read port value
	{
		char port;
		if (sscanf( (char*) inputBuffer, "fbreadioportlevel %c", &port) != 1) // input in HEX
		{
			printBadArgument();
		}
		else if (port == 'b')
		{
			readPortLevel(_PORTB_);
		}
		else if (port == 'd')
		{
			readPortLevel(_PORTD_);
		}
		else{
			printString("Supported ports - B and D");
			printBadArgument();
		}
	}
	//=====================================================================================================	
	else if (strlen( (char*) inputBuffer)!=0)
	{
		printUnknownCommand();
	}
	printPrompt();
	clearBuffer();
}

void printPrompt( void )
{
	printString("\r\n# ");
}

void printBadArgument( void )
{
	printString("Bad argument(s) ");
	printString( (char*) inputBuffer);
}

void printUnknownCommand( void )
{
		printString("Unknown command: ");
		printString( (char*) inputBuffer);
}

void printChannelHeader( void )
{
	printString("chan state mV mA");
}

/**
 * \brief Force read of channels
 * 
 * 
 * \return void
 */
void channelsSnapshot()
{
	for (int i = 0; i < CHANNEL_NUMBER; ++i)
	{
		if(!i2c_power_monitor_start_snapshot_measurement(&channels[i]))
		{
			channels[i].i2c_error = true;
		}
	}
	for (int i = 0; i < CHANNEL_NUMBER; ++i)
	{
		if (channels[i].i2c_error) continue;
		i2c_power_monitor_wait_for_snapshot(&channels[i]);
		i2c_power_monitor_read_voltage(&LOCAL_I2C, channels[i].i2c_addr, &(channels[i].voltage));
		i2c_power_monitor_read_current(&LOCAL_I2C, channels[i].i2c_addr, &(channels[i].current));
		i2c_power_monitor_read_power(&LOCAL_I2C, channels[i].i2c_addr, &(channels[i].power));
	}
}

/**
 * \brief Prints channel info
 * 
 * \param nr Channel number
 * 
 * \return void
 */
void printChannel( int nr )
{
	uint8_t status = 0;
	if (channels[nr].actual_state) status |= STATUS_ACTUAL_STATE;
	if (channels[nr].i2c_error) status |= STATUS_I2C_ERROR;
	if (channels[nr].alarm_pin0_status) status |= STATUS_ALARM_0;
	if (channels[nr].alarm_pin1_status) status |= STATUS_ALARM_1;
	if (channels[nr].overcurrent_alarm_status) status |= STATUS_OVERCURRENT;
	if (channels[nr].actual_state != channels[nr].state) status |= STATUS_UNEXPECTED;
	int voltage = (((channels[nr].negative) ? -1 : 1) * channels[nr].voltage);
	char str[60];
	sprintf(str, "\r\n%1d %02X %6d %5"PRIu32"", nr,
	status,
	voltage,
	channels[nr].current);
	printString(str);
}

void printAlarmMasks( void )
{
	char masks[15];
	sprintf(masks,"%02X ", getAlarmMask());
	printString(masks);
}

void printPowerStateWord( void )
{
	uint8_t word = readRegister(CPLD_ADDR_RELAY_STATUS);
	/*for (int i = 0; i < CHANNEL_NUMBER; ++i)
	{
		word |= ((channels[i].state?1:0) << i);
	}*/
	char str[4];
	sprintf(str, "%02X", word);
	printString(str);
}

void printLimits( void )
{
	for (int i = 0; i < CHANNEL_NUMBER; ++i)
	{
		char str[16];
		sprintf(str, "%1d %-5"PRIu32"\r\n", i, channels[i].current_limit);
		printString(str);
	}
}

void welcomeMessage( void )
{
	printPString("Fuse Relay Board\r\n");
	printPrompt();
}

void printPinAlarmMasks( void )
{
	uint8_t pin0 = readRegister(CPLD_ADDR_ALARMPIN0_MASK);
	uint8_t pin1 = readRegister(CPLD_ADDR_ALARMPIN1_MASK);
	char str[8];
	sprintf(str,"%02X %02X", pin0, pin1);
	printString(str);
}

void printVersion( void )
{
	char buf[100], id[64];
	sprintf(buf, "Firmware: %d.%d.%d CPLD %d", 
		FW_MAJOR_VERSION, FW_MINOR_VERSION, FW_REVISION, 
		readRegister(CPLD_ADDR_VERSION));
	printString(buf);
	uint8_t addr[6] = {0};
	if (i2c_read_uuid(addr))
	{
		sprintf(buf, "\r\nUUID: %02X-%02X-%02X-%02X-%02X-%02X", 
			addr[0], addr[1], addr[2], addr[3], 
			addr[4], addr[5]);
		printString(buf);
	}
	else
	{
		printString("\r\nUUID: 00-00-00-00-00-00");
	}
	eeprom_read_name(id);
	sprintf(buf, "\r\nBoard name: %s", id);
	printString(buf);
	eeprom_read_inv_no(id);
	sprintf(buf, "\r\nInventory number: %s", id);
	printString(buf);
}